from typing import Optional
from typing_extensions import Literal


from pydantic.main import BaseModel


class MaxChanges(BaseModel):
    enabled: bool
    limit: int


class Driving(BaseModel):
    type: Literal["driving"] = "driving"
    disable_border_crossing: Optional[bool] = None


class Walking(BaseModel):
    type: Literal["walking"] = "walking"


class Cycling(BaseModel):
    type: Literal["cycling"] = "cycling"


class Ferry(BaseModel):
    type: Literal["ferry", "cycling+ferry", "driving+ferry"] = "ferry"
    boarding_time: Optional[int] = None


class DrivingTrain(BaseModel):
    type: Literal["driving+train"] = "driving+train"
    pt_change_delay: Optional[int] = None
    driving_time_to_station: Optional[int] = None
    parking_time: Optional[int] = None
    walking_time: Optional[int] = None
    max_changes: Optional[MaxChanges] = None


class PublicTransport(BaseModel):
    type: Literal["public_transport", "train", "bus", "coach"] = "public_transport"
    pt_change_delay: Optional[int] = None
    walking_time: Optional[int] = None
    max_changes: Optional[MaxChanges] = None



from typing import List, Optional

from pydantic.main import BaseModel

from traveltimepy.dto.common import Fares, Route


class Property(BaseModel):
    travel_time: Optional[int]
    fares: Optional[Fares]
    distance: Optional[int]
    route: Optional[Route]


class Location(BaseModel):
    id: str
    properties: List[Property]


class RoutesResult(BaseModel):
    search_id: str
    locations: List[Location]
    unreachable: List[str]


class RoutesResponse(BaseModel):
    results: List[RoutesResult]



from typing import List, Optional

from pydantic import BaseModel


class TravelTime(BaseModel):
    min: int
    max: int
    mean: int
    median: int


class Properties(BaseModel):
    travel_time_reachable: Optional[TravelTime]
    travel_time_all: Optional[TravelTime]
    coverage: Optional[float]


class Zone(BaseModel):
    code: str
    properties: Properties


class PostcodesSectorsResult(BaseModel):
    search_id: str
    sectors: List[Zone]


class PostcodesDistrictsResult(BaseModel):
    search_id: str
    districts: List[Zone]


class PostcodesSectorsResponse(BaseModel):
    results: List[PostcodesSectorsResult]


class PostcodesDistrictsResponse(BaseModel):
    results: List[PostcodesDistrictsResult]



from datetime import datetime
from typing import List, Optional, Union

from pydantic.main import BaseModel

from traveltimepy.dto.common import Location, FullRange, Property
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.time_filter import TimeFilterResponse
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
)


class ArrivalSearch(BaseModel):
    id: str
    departure_location_ids: List[str]
    arrival_location_id: str
    arrival_time: datetime
    travel_time: int
    transportation: Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    properties: List[Property]
    range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: str
    arrival_location_ids: List[str]
    departure_location_id: str
    departure_time: datetime
    travel_time: int
    transportation: Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    properties: List[Property]
    range: Optional[FullRange] = None


class TimeFilterRequest(TravelTimeRequest[TimeFilterResponse]):
    locations: List[Location]
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeFilterRequest(
                locations=self.locations,
                departure_searches=departures,
                arrival_searches=arrivals,
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[TimeFilterResponse]) -> TimeFilterResponse:
        return TimeFilterResponse(
            results=flatten([response.results for response in responses])
        )



import typing
from datetime import datetime

from typing import List, Optional

from geojson_pydantic import FeatureCollection
from pydantic.main import BaseModel

from traveltimepy import (
    Coordinates,
    Range,
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
)
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.itertools import split, flatten


class DepartureSearch(BaseModel):
    id: str
    coords: Coordinates
    departure_time: datetime
    travel_time: int
    transportation: typing.Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    range: Optional[Range] = None


class ArrivalSearch(BaseModel):
    id: str
    coords: Coordinates
    arrival_time: datetime
    travel_time: int
    transportation: typing.Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    range: Optional[Range] = None


class TimeMapRequestGeojson(TravelTimeRequest[FeatureCollection]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[FeatureCollection]:
        return [
            FeatureCollection(
                type="FeatureCollection",
                departure_searches=departures,
                arrival_searches=arrivals,
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[FeatureCollection]) -> FeatureCollection:
        merged_features = flatten([response.features for response in responses])
        return FeatureCollection(type="FeatureCollection", features=merged_features)



import pytest

from traveltimepy import Coordinates
from traveltimepy.dto.requests.time_filter_proto import (
    ProtoTransportation,
    ProtoCountry,
)


@pytest.mark.asyncio
async def test_one_to_many(proto_sdk):
    travel_times = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
    )
    assert len(travel_times) == 2


@pytest.mark.asyncio
async def test_many_to_one(proto_sdk):
    travel_times = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        one_to_many=False
    )
    assert len(travel_times) == 2



import pytest
from datetime import datetime

from traveltimepy import Coordinates, PublicTransport


@pytest.mark.asyncio
async def test_departures(sdk):
    results = await sdk.postcodes_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        departure_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0


@pytest.mark.asyncio
async def test_arrivals(sdk):
    results = await sdk.postcodes_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        arrival_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0


@pytest.mark.asyncio
async def test_districts_departure(sdk):
    results = await sdk.postcodes_districts_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        departure_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0


@pytest.mark.asyncio
async def test_districts_arrival(sdk):
    results = await sdk.postcodes_districts_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        arrival_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0


@pytest.mark.asyncio
async def test_sectors_departure(sdk):
    results = await sdk.postcodes_sectors_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        departure_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0


@pytest.mark.asyncio
async def test_sectors_arrival(sdk):
    results = await sdk.postcodes_sectors_async(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        arrival_time=datetime.now(),
        transportation=PublicTransport(),
    )
    assert len(results) > 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TimeFilterFastRequest.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bTimeFilterFastRequest.proto\x12)com.igeolise.traveltime.rabbitmq.requests\"\"\n\x06\x43oords\x12\x0b\n\x03lat\x18\x01 \x01(\x02\x12\x0b\n\x03lng\x18\x02 \x01(\x02\"]\n\x0eTransportation\x12K\n\x04type\x18\x01 \x01(\x0e\x32=.com.igeolise.traveltime.rabbitmq.requests.TransportationType\"\x9f\x08\n\x15TimeFilterFastRequest\x12\x64\n\x10oneToManyRequest\x18\x01 \x01(\x0b\x32J.com.igeolise.traveltime.rabbitmq.requests.TimeFilterFastRequest.OneToMany\x12\x64\n\x10manyToOneRequest\x18\x02 \x01(\x0b\x32J.com.igeolise.traveltime.rabbitmq.requests.TimeFilterFastRequest.ManyToOne\x1a\x89\x03\n\tOneToMany\x12L\n\x11\x64\x65partureLocation\x18\x01 \x01(\x0b\x32\x31.com.igeolise.traveltime.rabbitmq.requests.Coords\x12\x16\n\x0elocationDeltas\x18\x02 \x03(\x11\x12Q\n\x0etransportation\x18\x03 \x01(\x0b\x32\x39.com.igeolise.traveltime.rabbitmq.requests.Transportation\x12P\n\x11\x61rrivalTimePeriod\x18\x04 \x01(\x0e\x32\x35.com.igeolise.traveltime.rabbitmq.requests.TimePeriod\x12\x12\n\ntravelTime\x18\x05 \x01(\x11\x12]\n\nproperties\x18\x06 \x03(\x0e\x32I.com.igeolise.traveltime.rabbitmq.requests.TimeFilterFastRequest.Property\x1a\x87\x03\n\tManyToOne\x12J\n\x0f\x61rrivalLocation\x18\x01 \x01(\x0b\x32\x31.com.igeolise.traveltime.rabbitmq.requests.Coords\x12\x16\n\x0elocationDeltas\x18\x02 \x03(\x11\x12Q\n\x0etransportation\x18\x03 \x01(\x0b\x32\x39.com.igeolise.traveltime.rabbitmq.requests.Transportation\x12P\n\x11\x61rrivalTimePeriod\x18\x04 \x01(\x0e\x32\x35.com.igeolise.traveltime.rabbitmq.requests.TimePeriod\x12\x12\n\ntravelTime\x18\x05 \x01(\x11\x12]\n\nproperties\x18\x06 \x03(\x0e\x32I.com.igeolise.traveltime.rabbitmq.requests.TimeFilterFastRequest.Property\"$\n\x08Property\x12\t\n\x05\x46\x41RES\x10\x00\x12\r\n\tDISTANCES\x10\x01*\xb8\x01\n\x12TransportationType\x12\x14\n\x10PUBLIC_TRANSPORT\x10\x00\x12\x0b\n\x07\x44RIVING\x10\x01\x12 \n\x1c\x44RIVING_AND_PUBLIC_TRANSPORT\x10\x02\x12\x15\n\x11\x44RIVING_AND_FERRY\x10\x03\x12\x0b\n\x07WALKING\x10\x04\x12\x15\n\x11WALKING_AND_FERRY\x10\x07\x12\x0b\n\x07\x43YCLING\x10\x05\x12\x15\n\x11\x43YCLING_AND_FERRY\x10\x06*!\n\nTimePeriod\x12\x13\n\x0fWEEKDAY_MORNING\x10\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'TimeFilterFastRequest_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSPORTATIONTYPE._serialized_start=1264
  _TRANSPORTATIONTYPE._serialized_end=1448
  _TIMEPERIOD._serialized_start=1450
  _TIMEPERIOD._serialized_end=1483
  _COORDS._serialized_start=74
  _COORDS._serialized_end=108
  _TRANSPORTATION._serialized_start=110
  _TRANSPORTATION._serialized_end=203
  _TIMEFILTERFASTREQUEST._serialized_start=206
  _TIMEFILTERFASTREQUEST._serialized_end=1261
  _TIMEFILTERFASTREQUEST_ONETOMANY._serialized_start=436
  _TIMEFILTERFASTREQUEST_ONETOMANY._serialized_end=829
  _TIMEFILTERFASTREQUEST_MANYTOONE._serialized_start=832
  _TIMEFILTERFASTREQUEST_MANYTOONE._serialized_end=1223
  _TIMEFILTERFASTREQUEST_PROPERTY._serialized_start=1225
  _TIMEFILTERFASTREQUEST_PROPERTY._serialized_end=1261
# @@protoc_insertion_point(module_scope)



from typing import Dict

from aiohttp import ClientSession, ClientResponse, BasicAuth, TCPConnector

from traveltimepy.TimeFilterFastResponse_pb2 import TimeFilterFastResponse
from traveltimepy.TimeFilterFastRequest_pb2 import TimeFilterFastRequest
from traveltimepy.dto.responses.time_filter_proto import TimeFilterProtoResponse
from traveltimepy.errors import ApiError


async def send_proto_async(
    url: str,
    headers: Dict[str, str],
    data: TimeFilterFastRequest,
    app_id: str,
    api_key: str,
) -> TimeFilterFastResponse:
    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        async with session.post(
            url=url,
            headers=headers,
            data=data.SerializeToString(),
            auth=BasicAuth(app_id, api_key),
        ) as resp:
            return await _process_response(resp)


async def _process_response(response: ClientResponse) -> TimeFilterProtoResponse:
    content = await response.read()
    if response.status != 200:
        msg = (
            f"Travel Time API proto request failed with error code: {response.status}\n"
        )
        raise ApiError(msg)
    else:
        response_body = TimeFilterFastResponse()
        response_body.ParseFromString(content)
        return TimeFilterProtoResponse(
            travel_times=response_body.properties.travelTimes[:]
        )



import pytest
from datetime import datetime

from traveltimepy import PublicTransport


@pytest.mark.asyncio
async def test_departures(sdk, locations):
    results = await sdk.routes_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=PublicTransport(),
        departure_time=datetime.now(),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_arrivals(sdk, locations):
    results = await sdk.routes_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=PublicTransport(),
        departure_time=datetime.now(),
    )
    assert len(results) == 2



from enum import Enum


class AcceptType(Enum):
    JSON = "application/json"
    BOUNDING_BOXES_JSON = "application/vnd.bounding-boxes+json"
    GEO_JSON = "application/geo+json"
    OCTET_STREAM = "application/octet-stream"



import asyncio
from dataclasses import dataclass
from typing import TypeVar, Type, Dict

from aiohttp import ClientSession, ClientResponse, TCPConnector
from pydantic.tools import parse_raw_as
from traveltimepy.dto.requests.request import TravelTimeRequest

from traveltimepy.dto.responses.error import ResponseError
from traveltimepy.errors import ApiError
from aiohttp_retry import RetryClient, ExponentialRetry
from aiolimiter import AsyncLimiter

T = TypeVar("T")
DEFAULT_SPLIT_SIZE = 10


@dataclass
class SdkParams:
    host: str
    proto_host: str
    limit_per_host: int
    rate_limit: int
    time_window: int
    retry_attempts: int


async def send_post_request_async(
    client: RetryClient,
    response_class: Type[T],
    url: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    rate_limit: AsyncLimiter,
) -> T:
    async with rate_limit:
        async with client.post(url=url, headers=headers, data=request.json()) as resp:
            return await _process_response(response_class, resp)


async def send_post_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    sdk_params: SdkParams,
) -> T:
    window_size = _window_size(sdk_params.rate_limit)
    async with ClientSession(
        connector=TCPConnector(ssl=False, limit_per_host=sdk_params.limit_per_host)
    ) as session:
        retry_options = ExponentialRetry(attempts=sdk_params.retry_attempts)
        async with RetryClient(
            client_session=session, retry_options=retry_options
        ) as client:
            rate_limit = AsyncLimiter(
                sdk_params.rate_limit // window_size, sdk_params.time_window
            )
            tasks = [
                send_post_request_async(
                    client,
                    response_class,
                    f"https://{sdk_params.host}/v4/{path}",
                    headers,
                    part,
                    rate_limit,
                )
                for part in request.split_searches(window_size)
            ]
            responses = await asyncio.gather(*tasks)
            return request.merge(responses)


async def send_post_geojson_async(
        response_class: Type[T],
        path: str,
        headers: Dict[str, str],
        request: TravelTimeRequest,
        sdk_params: SdkParams,
) -> T:
    window_size = _window_size(sdk_params.rate_limit)
    async with ClientSession(
            connector=TCPConnector(ssl=False, limit_per_host=sdk_params.limit_per_host)
    ) as session:
        retry_options = ExponentialRetry(attempts=sdk_params.retry_attempts)
        async with RetryClient(
                client_session=session, retry_options=retry_options
        ) as client:
            rate_limit = AsyncLimiter(
                sdk_params.rate_limit // window_size, sdk_params.time_window
            )
            tasks = [
                send_post_request_async(
                    client,
                    response_class,
                    f"https://{sdk_params.host}/v4/{path}",
                    headers,
                    request,
                    rate_limit,
                )
            ]
            responses = await asyncio.gather(*tasks)
            return request.merge(responses)


def _window_size(rate_limit: int):
    if rate_limit >= DEFAULT_SPLIT_SIZE:
        return DEFAULT_SPLIT_SIZE
    else:
        return rate_limit


async def send_get_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    sdk_params: SdkParams,
    params: Dict[str, str] = None,
) -> T:
    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        retry_options = ExponentialRetry(attempts=sdk_params.retry_attempts)
        async with RetryClient(
            client_session=session, retry_options=retry_options
        ) as client:
            async with client.get(
                url=f"https://{sdk_params.host}/v4/{path}",
                headers=headers,
                params=params,
            ) as resp:
                return await _process_response(response_class, resp)


async def _process_response(response_class: Type[T], response: ClientResponse) -> T:
    text = await response.text()
    if response.status != 200:
        parsed = parse_raw_as(ResponseError, text)
        msg = (
            f"Travel Time API request failed: {parsed.description}\n"
            f"Error code: {parsed.error_code}\n"
            f"Additional info: {parsed.additional_info}\n"
            f"<{parsed.documentation_link}>\n"
        )
        raise ApiError(msg)
    else:
        return parse_raw_as(response_class, text)



import pytest


@pytest.mark.asyncio
async def test_geocoding_search(sdk):
    response = await sdk.geocoding_async(
        query="Parliament square", limit=30, within_countries=["gb", "de"]
    )
    assert len(response.features) > 0
    assert len(response.features) < 31


@pytest.mark.asyncio
async def test_geocoding_reverse(sdk):
    response = await sdk.geocoding_reverse_async(lat=51.507281, lng=-0.132120)
    assert len(response.features) > 0



from datetime import datetime
from typing import List, Union, Optional

from pydantic import BaseModel

from traveltimepy.dto.common import Coordinates, Property, FullRange
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.postcodes import PostcodesResponse
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
)


class ArrivalSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    arrival_time: datetime
    transportation: Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    properties: List[Property]
    range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    departure_time: datetime
    transportation: Union[
        PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
    ]
    properties: List[Property]
    range: Optional[FullRange] = None


class PostcodesRequest(TravelTimeRequest[PostcodesResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            PostcodesRequest(departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[PostcodesResponse]) -> PostcodesResponse:
        return PostcodesResponse(
            results=sorted(
                flatten([response.results for response in responses]),
                key=lambda res: res.search_id,
            )
        )
